from dataclasses import dataclass, field
from datetime import datetime
from typing import Dict, List, Optional

import pydantic

from datahub.ingestion.source.sql.sql_common import SQLSourceReport


@dataclass
class BigQueryReport(SQLSourceReport):
    num_total_lineage_entries: Optional[int] = None
    num_skipped_lineage_entries_missing_data: Optional[int] = None
    num_skipped_lineage_entries_not_allowed: Optional[int] = None
    num_skipped_lineage_entries_sql_parser_failure: Optional[int] = None
    num_skipped_lineage_entries_other: Optional[int] = None
    num_total_log_entries: Optional[int] = None
    num_parsed_log_entires: Optional[int] = None
    num_total_audit_entries: Optional[int] = None
    num_parsed_audit_entires: Optional[int] = None
    bigquery_audit_metadata_datasets_missing: Optional[bool] = None
    lineage_metadata_entries: Optional[int] = None
    include_table_lineage: Optional[bool] = None
    use_date_sharded_audit_log_tables: Optional[bool] = None
    log_page_size: Optional[pydantic.PositiveInt] = None
    use_v2_audit_metadata: Optional[bool] = None
    use_exported_bigquery_audit_metadata: Optional[bool] = None
    window_start_time: Optional[datetime] = None
    window_end_time: Optional[datetime] = None
    log_entry_start_time: Optional[str] = None
    log_entry_end_time: Optional[str] = None
    audit_start_time: Optional[str] = None
    audit_end_time: Optional[str] = None
    upstream_lineage: Dict = field(default_factory=dict)
    partition_info: Dict[str, str] = field(default_factory=dict)
    table_metadata: Dict[str, List[str]] = field(default_factory=dict)
    profile_table_selection_criteria: Dict[str, str] = field(default_factory=dict)
    selected_profile_tables: Dict[str, List[str]] = field(default_factory=dict)
    invalid_partition_ids: Dict[str, str] = field(default_factory=dict)



from typing import Iterable, List

import pydantic

from datahub.emitter.mce_builder import (
    assertion_urn_to_key,
    container_urn_to_key,
    make_assertion_urn,
    make_container_urn,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityCheckpointStateBase,
)
from datahub.utilities.checkpoint_state_util import CheckpointStateUtil


class BaseSQLAlchemyCheckpointState(
    StaleEntityCheckpointStateBase["BaseSQLAlchemyCheckpointState"]
):
    """
    Base class for representing the checkpoint state for all SQLAlchemy based sources.
    Stores all tables and views being ingested and is used to remove any stale entities.
    Subclasses can define additional state as appropriate.
    """

    encoded_table_urns: List[str] = pydantic.Field(default_factory=list)
    encoded_view_urns: List[str] = pydantic.Field(default_factory=list)
    encoded_container_urns: List[str] = pydantic.Field(default_factory=list)
    encoded_assertion_urns: List[str] = pydantic.Field(default_factory=list)

    @classmethod
    def get_supported_types(cls) -> List[str]:
        return ["assertion", "container", "table", "view"]

    @staticmethod
    def _get_lightweight_repr(dataset_urn: str) -> str:
        """Reduces the amount of text in the URNs for smaller state footprint."""
        return CheckpointStateUtil.get_dataset_lightweight_repr(dataset_urn)

    @staticmethod
    def _get_container_lightweight_repr(container_urn: str) -> str:
        """Reduces the amount of text in the URNs for smaller state footprint."""
        key = container_urn_to_key(container_urn)
        assert key is not None
        return f"{key.guid}"

    @staticmethod
    def _get_container_urns_not_in(
        encoded_urns_1: List[str], encoded_urns_2: List[str]
    ) -> Iterable[str]:
        difference = CheckpointStateUtil.get_encoded_urns_not_in(
            encoded_urns_1, encoded_urns_2
        )
        for guid in difference:
            yield make_container_urn(guid)

    def _get_table_urns_not_in(
        self, checkpoint: "BaseSQLAlchemyCheckpointState"
    ) -> Iterable[str]:
        """Tables are mapped to DataHub dataset concept."""
        yield from CheckpointStateUtil.get_dataset_urns_not_in(
            self.encoded_table_urns, checkpoint.encoded_table_urns
        )

    def _get_view_urns_not_in(
        self, checkpoint: "BaseSQLAlchemyCheckpointState"
    ) -> Iterable[str]:
        """Views are mapped to DataHub dataset concept."""
        yield from CheckpointStateUtil.get_dataset_urns_not_in(
            self.encoded_view_urns, checkpoint.encoded_view_urns
        )

    def _get_assertion_urns_not_in(
        self, checkpoint: "BaseSQLAlchemyCheckpointState"
    ) -> Iterable[str]:
        """Tables are mapped to DataHub dataset concept."""
        diff = CheckpointStateUtil.get_encoded_urns_not_in(
            self.encoded_assertion_urns, checkpoint.encoded_assertion_urns
        )
        for assertion_id in diff:
            yield make_assertion_urn(assertion_id)

    def _add_table_urn(self, table_urn: str) -> None:
        self.encoded_table_urns.append(self._get_lightweight_repr(table_urn))

    def _add_assertion_urn(self, assertion_urn: str) -> None:
        key = assertion_urn_to_key(assertion_urn)
        assert key is not None
        self.encoded_assertion_urns.append(key.assertionId)

    def _add_view_urn(self, view_urn: str) -> None:
        self.encoded_view_urns.append(self._get_lightweight_repr(view_urn))

    def _add_container_guid(self, container_urn: str) -> None:
        self.encoded_container_urns.append(
            self._get_container_lightweight_repr(container_urn)
        )

    def add_checkpoint_urn(self, type: str, urn: str) -> None:
        assert type in self.get_supported_types()
        if type == "assertion":
            self._add_assertion_urn(urn)
        elif type == "container":
            self._add_container_guid(urn)
        elif type == "table":
            self._add_table_urn(urn)
        elif type == "view":
            self._add_view_urn(urn)

    def get_urns_not_in(
        self, type: str, other_checkpoint_state: "BaseSQLAlchemyCheckpointState"
    ) -> Iterable[str]:
        assert type in self.get_supported_types()
        if type == "assertion":
            yield from self._get_assertion_urns_not_in(other_checkpoint_state)
        if type == "container":
            yield from self._get_container_urns_not_in(
                self.encoded_container_urns,
                other_checkpoint_state.encoded_container_urns,
            )
        elif type == "table":
            yield from self._get_table_urns_not_in(other_checkpoint_state)
        elif type == "view":
            yield from self._get_view_urns_not_in(other_checkpoint_state)



# This module is written on the basis of below requirement
#     1) Entity absolute stat
#     2) Entity timeseries stat
#     3) Entity timeseries stat by user

import concurrent
import dataclasses
import datetime
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, cast

from looker_sdk.sdk.api31.models import Dashboard, LookWithQuery

import datahub.emitter.mce_builder as builder
from datahub.emitter.mce_builder import Aspect
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.source.looker import looker_common
from datahub.ingestion.source.looker.looker_common import (
    LookerDashboardSourceReport,
    LookerUser,
    LookerUserRegistry,
)
from datahub.ingestion.source.looker.looker_lib_wrapper import LookerAPI
from datahub.ingestion.source.looker.looker_query_model import (
    HistoryViewField,
    LookerExplore,
    LookerModel,
    LookerQuery,
    LookViewField,
    StrEnum,
    UserViewField,
    ViewField,
)
from datahub.metadata.schema_classes import (
    CalendarIntervalClass,
    ChangeTypeClass,
    ChartUsageStatisticsClass,
    ChartUserUsageCountsClass,
    DashboardUsageStatisticsClass,
    DashboardUserUsageCountsClass,
    TimeWindowSizeClass,
)

logger = logging.getLogger(__name__)


@dataclass
class ModelForUsage:
    id: Optional[str]


@dataclass
class LookerChartForUsage(ModelForUsage):
    view_count: Optional[int]

    @classmethod
    def from_chart(cls, chart: LookWithQuery) -> "LookerChartForUsage":
        return cls(id=str(chart.id) if chart.id else None, view_count=chart.view_count)


@dataclass
class LookerDashboardForUsage(ModelForUsage):
    view_count: Optional[int]
    favorite_count: Optional[int]
    last_viewed_at: Optional[int]
    looks: List[LookerChartForUsage] = dataclasses.field(default_factory=list)

    @classmethod
    def from_dashboard(cls, dashboard: Dashboard) -> "LookerDashboardForUsage":
        return cls(
            id=dashboard.id,
            view_count=dashboard.view_count,
            favorite_count=dashboard.favorite_count,
            last_viewed_at=round(dashboard.last_viewed_at.timestamp() * 1000)
            if dashboard.last_viewed_at
            else None,
            looks=[
                LookerChartForUsage.from_chart(e.look)
                for e in dashboard.dashboard_elements
                if e.look is not None
            ]
            if dashboard.dashboard_elements
            else [],
        )


@dataclass
class StatGeneratorConfig:
    looker_api_wrapper: LookerAPI
    looker_user_registry: LookerUserRegistry
    strip_user_ids_from_email: bool
    interval: str
    platform_name: str
    max_threads: int = 1


# QueryId and query_collection helps to return dummy responses in test cases
class QueryId(StrEnum):
    DASHBOARD_PER_DAY_USAGE_STAT = "counts_per_day_per_dashboard"
    DASHBOARD_PER_USER_PER_DAY_USAGE_STAT = "counts_per_day_per_user_per_dashboard"
    LOOK_PER_DAY_USAGE_STAT = "counts_per_day_per_look"
    LOOK_PER_USER_PER_DAY_USAGE_STAT = "counts_per_day_per_user_per_look"


query_collection: Dict[QueryId, LookerQuery] = {
    QueryId.DASHBOARD_PER_DAY_USAGE_STAT: LookerQuery(
        model=LookerModel.SYSTEM_ACTIVITY,
        explore=LookerExplore.HISTORY,
        fields=[
            HistoryViewField.HISTORY_DASHBOARD_ID,
            HistoryViewField.HISTORY_CREATED_DATE,
            HistoryViewField.HISTORY_DASHBOARD_USER,
            HistoryViewField.HISTORY_DASHBOARD_RUN_COUNT,
        ],
    ),
    QueryId.DASHBOARD_PER_USER_PER_DAY_USAGE_STAT: LookerQuery(
        model=LookerModel.SYSTEM_ACTIVITY,
        explore=LookerExplore.HISTORY,
        fields=[
            HistoryViewField.HISTORY_CREATED_DATE,
            HistoryViewField.HISTORY_DASHBOARD_ID,
            HistoryViewField.HISTORY_DASHBOARD_RUN_COUNT,
            UserViewField.USER_ID,
        ],
    ),
    QueryId.LOOK_PER_DAY_USAGE_STAT: LookerQuery(
        model=LookerModel.SYSTEM_ACTIVITY,
        explore=LookerExplore.HISTORY,
        fields=[
            HistoryViewField.HISTORY_CREATED_DATE,
            HistoryViewField.HISTORY_COUNT,
            LookViewField.LOOK_ID,
        ],
        filters={
            LookViewField.LOOK_ID: "NOT NULL",
        },
    ),
    QueryId.LOOK_PER_USER_PER_DAY_USAGE_STAT: LookerQuery(
        model=LookerModel.SYSTEM_ACTIVITY,
        explore=LookerExplore.HISTORY,
        fields=[
            HistoryViewField.HISTORY_CREATED_DATE,
            HistoryViewField.HISTORY_COUNT,
            LookViewField.LOOK_ID,
            UserViewField.USER_ID,
        ],
        filters={
            LookViewField.LOOK_ID: "NOT NULL",
        },
    ),
}


class BaseStatGenerator(ABC):
    """
    A Template class to encapsulate looker stat processing
    """

    def __init__(
        self,
        config: StatGeneratorConfig,
        looker_models: List[ModelForUsage],
        report: LookerDashboardSourceReport,
    ):
        self.config = config
        self.looker_models = looker_models
        # Later it will help to find out for what are the looker entities from query result
        self.id_vs_model: Dict[str, ModelForUsage] = {
            self.get_id(looker_object): looker_object for looker_object in looker_models
        }
        self.post_filter = len(self.looker_models) > 100
        self.report = report

    @abstractmethod
    def get_stats_generator_name(self) -> str:
        """All implementing classes should provide a name identifying them"""
        pass

    @abstractmethod
    def get_entity_timeseries_query(self) -> LookerQuery:
        pass

    @abstractmethod
    def get_entity_user_timeseries_query(self) -> LookerQuery:
        pass

    @abstractmethod
    def get_filter(self) -> Dict[ViewField, str]:
        pass

    @abstractmethod
    def to_entity_absolute_stat_aspect(self, looker_object: ModelForUsage) -> Aspect:
        pass

    @abstractmethod
    def to_entity_timeseries_stat_aspect(self, row: Dict) -> Aspect:
        pass

    @abstractmethod
    def get_entity_stat_key(self, row: Dict) -> Tuple[str, str]:
        pass

    @abstractmethod
    def _get_mcp_attributes(self, model: ModelForUsage) -> Dict:
        pass

    @abstractmethod
    def append_user_stat(
        self, entity_stat_aspect: Aspect, user: LookerUser, row: Dict
    ) -> None:
        pass

    @abstractmethod
    def get_id(self, looker_object: ModelForUsage) -> str:
        pass

    @abstractmethod
    def get_id_from_row(self, row: dict) -> str:
        pass

    def create_mcp(
        self, model: ModelForUsage, aspect: Aspect
    ) -> MetadataChangeProposalWrapper:
        return MetadataChangeProposalWrapper(
            aspect=aspect,
            **self._get_mcp_attributes(model=model),
        )

    def _round_time(self, date_time: str) -> int:
        return round(
            datetime.datetime.strptime(date_time, "%Y-%m-%d")
            .replace(tzinfo=datetime.timezone.utc)
            .timestamp()
            * 1000
        )

    def _get_user_identifier(self, row: Dict) -> int:
        return row[UserViewField.USER_ID]

    def _process_entity_timeseries_rows(
        self, rows: List[Dict]
    ) -> Dict[Tuple[str, str], Aspect]:
        # Convert Looker entity stat i.e. rows to DataHub stat aspect
        entity_stat_aspect: Dict[Tuple[str, str], Aspect] = {}

        for row in rows:
            logger.debug(row)
            entity_stat_aspect[
                self.get_entity_stat_key(row)
            ] = self.to_entity_timeseries_stat_aspect(row)

        return entity_stat_aspect

    def _process_absolute_aspect(self) -> List[Tuple[ModelForUsage, Aspect]]:
        aspects: List[Tuple[ModelForUsage, Aspect]] = []
        for looker_object in self.looker_models:
            aspects.append(
                (looker_object, self.to_entity_absolute_stat_aspect(looker_object))
            )

        return aspects

    def _fill_user_stat_aspect(
        self,
        entity_usage_stat: Dict[Tuple[str, str], Aspect],
        user_wise_rows: List[Dict],
    ) -> Iterable[Tuple[ModelForUsage, Aspect]]:
        logger.debug("Entering fill user stat aspect")

        # We first resolve all the users using a threadpool to warm up the cache
        user_ids = set([self._get_user_identifier(row) for row in user_wise_rows])
        start_time = datetime.datetime.now()
        with concurrent.futures.ThreadPoolExecutor(
            max_workers=self.config.max_threads
        ) as async_executor:
            user_futures = {
                async_executor.submit(
                    self.config.looker_user_registry.get_by_id, user_id
                ): 1
                for user_id in user_ids
            }
            for future in concurrent.futures.as_completed(user_futures):
                logger.debug("User result: %s", future.result())
                del user_futures[future]

        user_resolution_latency = datetime.datetime.now() - start_time
        logger.debug(
            f"Resolved {len(user_ids)} in {user_resolution_latency.total_seconds()} seconds"
        )
        self.report.report_user_resolution_latency(
            self.get_stats_generator_name(), user_resolution_latency
        )

        for row in user_wise_rows:
            # Confirm looker object was given for stat generation
            looker_object = self.id_vs_model.get(self.get_id_from_row(row))
            if looker_object is None:
                logger.warning(
                    "Looker object with id({}) was not register with stat generator".format(
                        self.get_id_from_row(row)
                    )
                )
                continue

            # Confirm we do have user with user id
            user: Optional[LookerUser] = self.config.looker_user_registry.get_by_id(
                self._get_user_identifier(row)
            )
            if user is None:
                logger.warning(
                    f"Unable to resolve user with id {self._get_user_identifier(row)}, skipping"
                )
                continue

            # Confirm for the user row (entity + time) the entity stat is present for the same time
            entity_stat_aspect: Optional[Aspect] = entity_usage_stat.get(
                self.get_entity_stat_key(row)
            )
            if entity_stat_aspect is None:
                logger.warning(
                    f"entity stat is not found for the row = {row}, in entity_stat = {entity_usage_stat}"
                )
                logger.warning(
                    "entity stat is not found for the user stat key = {}".format(
                        self.get_entity_stat_key(row)
                    )
                )
                continue
            self.append_user_stat(entity_stat_aspect, user, row)

        logger.debug("Starting to yield answers for user-wise counts")

        for (id, _), aspect in entity_usage_stat.items():
            yield self.id_vs_model[id], aspect

    def _execute_query(self, query: LookerQuery, query_name: str) -> List[Dict]:
        rows = []
        try:
            start_time = datetime.datetime.now()
            rows = self.config.looker_api_wrapper.execute_query(
                write_query=query.to_write_query()
            )
            end_time = datetime.datetime.now()

            logger.debug(
                f"{self.get_stats_generator_name()}: Retrieved {len(rows)} rows in {(end_time - start_time).total_seconds()} seconds"
            )
            self.report.report_query_latency(
                f"{self.get_stats_generator_name()}:{query_name}", end_time - start_time
            )
            if self.post_filter:
                logger.debug("post filtering")
                rows = [r for r in rows if self.get_id_from_row(r) in self.id_vs_model]
                logger.debug("Filtered down to %d rows", len(rows))
        except Exception as e:
            logger.warning(f"Failed to execute {query_name} query", e)

        return rows

    def _append_filters(self, query: LookerQuery) -> LookerQuery:
        query.filters.update(
            {HistoryViewField.HISTORY_CREATED_DATE: self.config.interval}
        )
        if not self.post_filter:
            query.filters.update(self.get_filter())
        return query

    def generate_usage_stat_mcps(self) -> Iterable[MetadataChangeProposalWrapper]:

        # No looker entities available to process stat generation
        if len(self.looker_models) == 0:
            return

        # yield absolute stat for looker entities
        for looker_object, aspect in self._process_absolute_aspect():  # type: ignore
            yield self.create_mcp(looker_object, aspect)

        # Execute query and process the raw json which contains stat information
        entity_query_with_filters: LookerQuery = self._append_filters(
            self.get_entity_timeseries_query()
        )
        entity_rows: List[Dict] = self._execute_query(
            entity_query_with_filters, "entity_query"
        )
        entity_usage_stat: Dict[
            Tuple[str, str], Any
        ] = self._process_entity_timeseries_rows(
            entity_rows
        )  # Any type to pass mypy unbound Aspect type error

        user_wise_query_with_filters: LookerQuery = self._append_filters(
            self.get_entity_user_timeseries_query()
        )
        user_wise_rows = self._execute_query(user_wise_query_with_filters, "user_query")
        # yield absolute stat for entity
        for looker_object, aspect in self._fill_user_stat_aspect(
            entity_usage_stat, user_wise_rows
        ):
            yield self.create_mcp(looker_object, aspect)


class DashboardStatGenerator(BaseStatGenerator):
    def __init__(
        self,
        config: StatGeneratorConfig,
        looker_dashboards: List[LookerDashboardForUsage],
        report: LookerDashboardSourceReport,
    ):
        super().__init__(
            config,
            looker_models=cast(List[ModelForUsage], looker_dashboards),
            report=report,
        )
        self.report = report
        self.report.report_dashboards_scanned_for_usage(len(looker_dashboards))

    def get_stats_generator_name(self) -> str:
        return "DashboardStats"

    def get_filter(self) -> Dict[ViewField, str]:
        return {
            HistoryViewField.HISTORY_DASHBOARD_ID: ",".join(
                [
                    looker_dashboard.id
                    for looker_dashboard in self.looker_models
                    if looker_dashboard.id is not None
                ]
            )
        }

    def get_id(self, looker_object: ModelForUsage) -> str:
        if looker_object.id is None:
            raise ValueError("Looker dashboard id is None")
        return looker_object.id

    def get_id_from_row(self, row: dict) -> str:
        return str(row[HistoryViewField.HISTORY_DASHBOARD_ID])

    def get_entity_stat_key(self, row: Dict) -> Tuple[str, str]:
        self.report.dashboards_with_activity.add(
            row[HistoryViewField.HISTORY_DASHBOARD_ID]
        )
        return (
            row[HistoryViewField.HISTORY_DASHBOARD_ID],
            row[HistoryViewField.HISTORY_CREATED_DATE],
        )

    def _get_mcp_attributes(self, model: ModelForUsage) -> Dict:
        dashboard: Dashboard = cast(Dashboard, model)
        if dashboard is None or dashboard.id is None:  # to pass mypy lint
            return {}

        return {
            "entityUrn": builder.make_dashboard_urn(
                self.config.platform_name,
                looker_common.get_urn_looker_dashboard_id(dashboard.id),
            ),
            "entityType": "dashboard",
            "changeType": ChangeTypeClass.UPSERT,
            "aspectName": "dashboardUsageStatistics",
        }

    def to_entity_absolute_stat_aspect(self, looker_object: ModelForUsage) -> Aspect:
        looker_dashboard: LookerDashboardForUsage = cast(
            LookerDashboardForUsage, looker_object
        )
        if looker_dashboard.view_count:
            self.report.dashboards_with_activity.add(str(looker_dashboard.id))
        return cast(
            Aspect,
            DashboardUsageStatisticsClass(
                timestampMillis=round(datetime.datetime.now().timestamp() * 1000),
                favoritesCount=looker_dashboard.favorite_count,
                viewsCount=looker_dashboard.view_count,
                lastViewedAt=looker_dashboard.last_viewed_at,
            ),
        )

    def get_entity_timeseries_query(self) -> LookerQuery:
        return query_collection[QueryId.DASHBOARD_PER_DAY_USAGE_STAT]

    def get_entity_user_timeseries_query(self) -> LookerQuery:
        return query_collection[QueryId.DASHBOARD_PER_USER_PER_DAY_USAGE_STAT]

    def to_entity_timeseries_stat_aspect(self, row: dict) -> Aspect:
        self.report.dashboards_with_activity.add(
            row[HistoryViewField.HISTORY_DASHBOARD_ID]
        )
        return cast(
            Aspect,
            DashboardUsageStatisticsClass(
                timestampMillis=self._round_time(
                    row[HistoryViewField.HISTORY_CREATED_DATE]
                ),
                eventGranularity=TimeWindowSizeClass(unit=CalendarIntervalClass.DAY),
                uniqueUserCount=row[HistoryViewField.HISTORY_DASHBOARD_USER],
                executionsCount=row[HistoryViewField.HISTORY_DASHBOARD_RUN_COUNT],
            ),
        )

    def append_user_stat(
        self, entity_stat_aspect: Aspect, user: LookerUser, row: Dict
    ) -> None:
        dashboard_stat_aspect: DashboardUsageStatisticsClass = cast(
            DashboardUsageStatisticsClass, entity_stat_aspect
        )

        if dashboard_stat_aspect.userCounts is None:
            dashboard_stat_aspect.userCounts = []

        user_urn: Optional[str] = user.get_urn(self.config.strip_user_ids_from_email)

        if user_urn is None:
            logger.warning("user_urn not found for the user {}".format(user))
            return

        dashboard_stat_aspect.userCounts.append(
            DashboardUserUsageCountsClass(
                user=user_urn,
                executionsCount=row[HistoryViewField.HISTORY_DASHBOARD_RUN_COUNT],
                usageCount=row[HistoryViewField.HISTORY_DASHBOARD_RUN_COUNT],
                userEmail=user.email,
            )
        )


class LookStatGenerator(BaseStatGenerator):
    def __init__(
        self,
        config: StatGeneratorConfig,
        looker_looks: List[LookerChartForUsage],
        report: LookerDashboardSourceReport,
    ):
        super().__init__(
            config,
            looker_models=cast(List[ModelForUsage], looker_looks),
            report=report,
        )
        self.report = report
        report.report_charts_scanned_for_usage(len(looker_looks))

    def get_stats_generator_name(self) -> str:
        return "ChartStats"

    def get_filter(self) -> Dict[ViewField, str]:
        return {
            LookViewField.LOOK_ID: ",".join(
                [look.id for look in self.looker_models if look.id is not None]
            )
        }

    def get_id(self, looker_object: ModelForUsage) -> str:
        look: LookerChartForUsage = cast(LookerChartForUsage, looker_object)
        if (
            look.id is None
        ):  # This condition never become true, this check is to pass the mypy lint error
            raise ValueError("Looker look model id is None")
        return str(look.id)

    def get_id_from_row(self, row: dict) -> str:
        return str(row[LookViewField.LOOK_ID])

    def get_entity_stat_key(self, row: Dict) -> Tuple[str, str]:
        return (
            str(row[LookViewField.LOOK_ID]),
            row[HistoryViewField.HISTORY_CREATED_DATE],
        )

    def _get_mcp_attributes(self, model: ModelForUsage) -> Dict:
        look: LookerChartForUsage = cast(LookerChartForUsage, model)
        if look is None or look.id is None:
            return {}

        return {
            "entityUrn": builder.make_chart_urn(
                self.config.platform_name,
                looker_common.get_urn_looker_element_id(str(look.id)),
            ),
            "entityType": "chart",
            "changeType": ChangeTypeClass.UPSERT,
            "aspectName": "chartUsageStatistics",
        }

    def to_entity_absolute_stat_aspect(self, looker_object: ModelForUsage) -> Aspect:
        looker_look: LookerChartForUsage = cast(LookerChartForUsage, looker_object)
        assert looker_look.id
        if looker_look.view_count:
            self.report.charts_with_activity.add(looker_look.id)

        return cast(
            Aspect,
            ChartUsageStatisticsClass(
                timestampMillis=round(datetime.datetime.now().timestamp() * 1000),
                viewsCount=looker_look.view_count,
            ),
        )

    def get_entity_timeseries_query(self) -> LookerQuery:
        return query_collection[QueryId.LOOK_PER_DAY_USAGE_STAT]

    def get_entity_user_timeseries_query(self) -> LookerQuery:
        return query_collection[QueryId.LOOK_PER_USER_PER_DAY_USAGE_STAT]

    def to_entity_timeseries_stat_aspect(self, row: dict) -> Aspect:
        self.report.charts_with_activity.add(str(row[LookViewField.LOOK_ID]))

        return cast(
            Aspect,
            ChartUsageStatisticsClass(
                timestampMillis=self._round_time(
                    row[HistoryViewField.HISTORY_CREATED_DATE]
                ),
                eventGranularity=TimeWindowSizeClass(unit=CalendarIntervalClass.DAY),
                viewsCount=row[HistoryViewField.HISTORY_COUNT],
            ),
        )

    def append_user_stat(
        self, entity_stat_aspect: Aspect, user: LookerUser, row: Dict
    ) -> None:
        chart_stat_aspect: ChartUsageStatisticsClass = cast(
            ChartUsageStatisticsClass, entity_stat_aspect
        )

        if chart_stat_aspect.userCounts is None:
            chart_stat_aspect.userCounts = []

        user_urn: Optional[str] = user.get_urn(self.config.strip_user_ids_from_email)

        if user_urn is None:
            logger.warning("user_urn not found for the user {}".format(user))
            return

        chart_stat_aspect.userCounts.append(
            ChartUserUsageCountsClass(
                user=user_urn,
                viewsCount=row[HistoryViewField.HISTORY_COUNT],
            )
        )


class SupportedStatEntity(Enum):
    DASHBOARD = "dashboard"
    CHART = "chart"


# type_ is because of type is builtin identifier
def create_stat_entity_generator(
    type_: SupportedStatEntity, config: StatGeneratorConfig
) -> Callable[[List[ModelForUsage], LookerDashboardSourceReport], BaseStatGenerator]:
    # Wrapper function to defer creation of actual entities
    # config is generally available at the startup, however entities may get created later during processing
    def create_dashboard_stat_generator(
        looker_dashboards: List[LookerDashboardForUsage],
        report: LookerDashboardSourceReport,
    ) -> BaseStatGenerator:
        logger.debug(
            "Number of dashboard received for stat processing = {}".format(
                len(looker_dashboards)
            )
        )
        return DashboardStatGenerator(
            config=config, looker_dashboards=looker_dashboards, report=report
        )

    def create_chart_stat_generator(
        looker_looks: List[LookerChartForUsage], report: LookerDashboardSourceReport
    ) -> BaseStatGenerator:
        logger.debug(
            "Number of looks received for stat processing = {}".format(
                len(looker_looks)
            )
        )
        return LookStatGenerator(
            config=config, looker_looks=looker_looks, report=report
        )

    stat_entities_generator = {
        SupportedStatEntity.DASHBOARD: create_dashboard_stat_generator,
        SupportedStatEntity.CHART: create_chart_stat_generator,
    }

    return stat_entities_generator[type_]  # type: ignore



from typing import Optional

from deltalake import DeltaTable, PyDeltaTableError

from datahub.ingestion.source.delta_lake.config import DeltaLakeSourceConfig


def read_delta_table(
    path: str, delta_lake_config: DeltaLakeSourceConfig
) -> Optional[DeltaTable]:
    delta_table = None
    try:
        opts = {}
        if delta_lake_config.is_s3:
            if (
                delta_lake_config.s3 is not None
                and delta_lake_config.s3.aws_config is not None
            ):
                creds = delta_lake_config.s3.aws_config.get_credentials()
                opts = {
                    "AWS_ACCESS_KEY_ID": creds.get("aws_access_key_id", ""),
                    "AWS_SECRET_ACCESS_KEY": creds.get("aws_secret_access_key", ""),
                }
        delta_table = DeltaTable(path, storage_options=opts)

    except PyDeltaTableError as e:
        if "Not a Delta table" not in str(e):
            raise e
    return delta_table


def get_file_count(delta_table: DeltaTable) -> int:
    return len(delta_table.files())



import typing
from typing import Any, Dict

import pydantic
from pydantic.fields import Field
from sqlalchemy import create_engine, inspect
from sqlalchemy.engine.reflection import Inspector

from datahub.configuration.common import AllowDenyPattern
from datahub.emitter.mcp_builder import PlatformKey
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.sql.sql_common import (
    BasicSQLAlchemyConfig,
    SQLAlchemySource,
    logger,
    make_sqlalchemy_uri,
)


class TwoTierSQLAlchemyConfig(BasicSQLAlchemyConfig):

    database_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for databases to filter in ingestion.",
    )
    schema_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Deprecated in favour of database_pattern. Regex patterns for schemas to filter in ingestion. "
        "Specify regex to only match the schema name. e.g. to match all tables in schema analytics, "
        "use the regex 'analytics'",
    )

    @pydantic.root_validator()
    def ensure_profiling_pattern_is_passed_to_profiling(
        cls, values: Dict[str, Any]
    ) -> Dict[str, Any]:
        allow_all_pattern = AllowDenyPattern.allow_all()
        schema_pattern = values.get("schema_pattern")
        database_pattern = values.get("database_pattern")
        if (
            database_pattern == allow_all_pattern
            and schema_pattern != allow_all_pattern
        ):
            logger.warning(
                "Updating 'database_pattern' to 'schema_pattern'. Please stop using deprecated "
                "'schema_pattern'. Use 'database_pattern' instead. "
            )
            values["database_pattern"] = schema_pattern
        return values

    def get_sql_alchemy_url(
        self,
        uri_opts: typing.Optional[typing.Dict[str, typing.Any]] = None,
        current_db: typing.Optional[str] = None,
    ) -> str:
        return self.sqlalchemy_uri or make_sqlalchemy_uri(
            self.scheme,  # type: ignore
            self.username,
            self.password.get_secret_value() if self.password else None,
            self.host_port,  # type: ignore
            current_db if current_db else self.database,
            uri_opts=uri_opts,
        )


class TwoTierSQLAlchemySource(SQLAlchemySource):
    def __init__(self, config, ctx, platform):
        super().__init__(config, ctx, platform)
        self.current_database = None
        self.config: TwoTierSQLAlchemyConfig = config

    def get_parent_container_key(self, db_name: str, schema: str) -> PlatformKey:
        return self.gen_database_key(db_name)

    def get_allowed_schemas(
        self, inspector: Inspector, db_name: str
    ) -> typing.Iterable[str]:
        # This method returns schema names but for 2 tier databases there is no schema layer at all hence passing
        # dbName itself as an allowed schema
        yield db_name

    def get_inspectors(self):
        # This method can be overridden in the case that you want to dynamically
        # run on multiple databases.
        url = self.config.get_sql_alchemy_url()
        logger.debug(f"sql_alchemy_url={url}")
        engine = create_engine(url, **self.config.options)
        with engine.connect() as conn:
            inspector = inspect(conn)
            if self.config.database and self.config.database != "":
                databases = [self.config.database]
            else:
                databases = inspector.get_schema_names()
            for db in databases:
                if self.config.database_pattern.allowed(db):
                    url = self.config.get_sql_alchemy_url(current_db=db)
                    inspector = inspect(
                        create_engine(url, **self.config.options).connect()
                    )
                    self.current_database = db
                    yield inspector

    def gen_schema_containers(
        self, schema: str, db_name: str
    ) -> typing.Iterable[MetadataWorkUnit]:
        return []



from datahub.metadata.schema_classes import ASPECT_CLASSES

ASPECT_MAP = {
    AspectClass.get_aspect_name(): AspectClass for AspectClass in ASPECT_CLASSES
}

TIMESERIES_ASPECT_MAP = {
    name: klass
    for name, klass in ASPECT_MAP.items()
    if klass.get_aspect_type() == "timeseries"
}



import time
from datetime import datetime, timezone


def get_current_time_in_seconds() -> int:
    return int(time.time())


def get_datetime_from_ts_millis_in_utc(ts_millis: int) -> datetime:
    return datetime.fromtimestamp(ts_millis / 1000, tz=timezone.utc)


def datetime_to_ts_millis(dt: datetime) -> int:
    return int(
        round(dt.timestamp() * 1000),
    )
